#!/usr/bin/env python
"""
Sample Barebones Pandora Player

This is a very simple Pandora player that streams music from Pandora. It
requires mpg123 to function. No songs are downloaded, they are streamed
directly from Pandora's servers.
"""
import os
import sys

from pandora import APIClient
from pydora.mpg123 import Player
from .utils import Colors, Screen


class PlayerApp:

    CMD_MAP = {
        'n': ('play next song', 'skip_song'),
        'p': ('pause/resume song', 'pause_song'),
        's': ('stop playing station', 'stop_station'),
        'd': ('dislike song', 'dislike_song'),
        'u': ('like song', 'like_song'),
        'b': ('bookmark song', 'bookmark_song'),
        'a': ('bookmark artist', 'bookmark_artist'),
        'S': ('sleep song for 30 days', 'sleep_song'),
        'Q': ('quit player', 'quit'),
        '?': ('display this help', 'help'),
    }

    def __init__(self):
        self.client = None
        self.player = Player(self, sys.stdin)

    @property
    def config_path(self):
        """Find the config file

        Config file exists in either ~/.pydora.cfg or is pointed to by an
        environment variable PYDORA_CFG.
        """
        path = os.path.expanduser(
            os.environ.get('PYDORA_CFG', '~/.pydora.cfg'))

        if not os.path.exists(path):
            Screen.print_error('No settings at {!r}'.format(path))
            sys.exit(1)

        return path

    def station_selection_menu(self):
        """Format a station menu and make the user select a station
        """
        Screen.clear()

        for i, s in enumerate(self.stations):
            i = '{:>3}'.format(i)
            print('{}: {}'.format(Colors.yellow(i), s.name))

        return self.stations[Screen.get_integer('Station: ')]

    def play(self, song):
        """Play callback
        """
        print(u'{} by {}'.format(Colors.cyan(song.song_name),
            Colors.yellow(song.artist_name)))

    def skip_song(self, song):
        self.player.stop()

    def pause_song(self, song):
        self.player.pause()

    def stop_station(self, song):
        self.player.end_station()

    def dislike_song(self, song):
        song.thumbs_down()
        Screen.print_success('Track disliked')
        self.player.stop()

    def like_song(self, song):
        song.thumbs_up()
        Screen.print_success('Track liked')

    def bookmark_song(self, song):
        song.bookmark_song()
        Screen.print_success('Bookmarked song')

    def bookmark_artist(self, song):
        song.bookmark_artist()
        Screen.print_success('Bookmarked artist')

    def sleep_song(self, song):
        song.sleep()
        Screen.print_success('Song will not be played for 30 days')
        self.player.stop()

    def quit(self, song):
        self.player.end_station()
        sys.exit(0)

    def help(self, song):
        print("")
        print("\n".join([
            "\t{} - {}".format(k, v[0]) for k, v in self.CMD_MAP.items()
        ]))
        print("")

    def input(self, input, song):
        """Input callback, handles key presses
        """
        try:
            cmd = getattr(self, self.CMD_MAP[input][1])
        except (IndexError, KeyError):
            return Screen.print_error('Invalid command!')

        cmd(song)

    def pre_poll(self):
        Screen.set_echo(False)

    def post_poll(self):
        Screen.set_echo(True)

    def run(self):
        self.client = APIClient.from_config_file(self.config_path)
        self.stations = self.client.get_station_list()

        while True:
            try:
                station = self.station_selection_menu()
                self.player.play_station(station)
            except KeyboardInterrupt:
                sys.exit(0)


def main():
    PlayerApp().run()


if __name__ == '__main__':
    main()



from datetime import datetime
from collections import namedtuple


def with_metaclass(meta, *bases):
    return meta("NewBase", bases, {})


class Field(namedtuple("Field", ["field", "default", "formatter"])):

    def __new__(cls, field, default=None, formatter=None):
        return super(Field, cls).__new__(cls, field, default, formatter)


class ModelMetaClass(type):

    def __new__(cls, name, parents, dct):
        dct["_fields"] = fields = {}
        new_dct = dct.copy()

        for key, val in dct.items():
            if isinstance(val, Field):
                fields[key] = val
                del new_dct[key]

        return super(ModelMetaClass, cls).__new__(cls, name, parents, new_dct)


class PandoraModel(with_metaclass(ModelMetaClass, object)):

    @staticmethod
    def json_to_date(api_client, data):
        return datetime.utcfromtimestamp(data["time"] / 1000)

    def __init__(self, api_client):
        self._api_client = api_client

        try:
            safe_types = (type(None), basestring, int, bool)
        except NameError:
            safe_types = (type(None), str, bytes, int, bool)

        for key, value in self._fields.items():
            default = value.default

            if not isinstance(default, safe_types):
                default = type(default)()

            setattr(self, key, default)

    @staticmethod
    def populate_fields(api_client, instance, data):
        for key, value in instance.__class__._fields.items():
            if key.startswith("__"):
                continue

            newval = data.get(value.field, value.default)

            if newval and value.formatter:
                newval = value.formatter(api_client, newval)

            setattr(instance, key, newval)

    @classmethod
    def from_json(cls, api_client, data):
        self = cls(api_client)
        PandoraModel.populate_fields(api_client, self, data)
        return self

    @classmethod
    def from_json_list(cls, api_client, data):
        return [cls.from_json(api_client, item) for item in data]

    def __repr__(self):
        output = ", ".join([
            "=".join((key, repr(getattr(self, key))))
            for key in self._fields.keys()])

        return "{}({})".format(self.__class__.__name__, output)


class PandoraListModel(PandoraModel, list):

    __list_key__ = None
    __list_model__ = None

    @classmethod
    def from_json(cls, api_client, data):
        self = cls(api_client)
        PandoraModel.populate_fields(api_client, self, data)

        for station in data[cls.__list_key__]:
            self.append(cls.__list_model__.from_json(api_client, station))

        return self



from .. import BaseAPIClient, URLTester
from . import Field, PandoraModel, PandoraListModel


class Station(PandoraModel):

    can_add_music = Field("allowAddMusic", False)
    can_delete = Field("allowDelete", True)
    can_rename = Field("allowRename", True)
    is_shared = Field("isShared", False)

    art_url = Field("artUrl")
    date_created = Field("dateCreated", formatter=PandoraModel.json_to_date)
    detail_url = Field("stationDetailUrl")
    id = Field("stationId")
    name = Field("stationName")
    sharing_url = Field("stationSharingUrl")
    token = Field("stationToken")

    genre = Field("genre", [])
    quickmix_stations = Field("quickMixStationIds", [])

    def get_playlist(self):
        return iter(self._api_client.get_playlist(self.token))


class StationList(PandoraListModel):

    checksum = Field("checksum")

    __list_key__ = "stations"
    __list_model__ = Station

    def has_changed(self):
        checksum = self._api_client.get_station_list_checksum()
        return checksum != self.checksum


class PlaylistItem(PandoraModel):

    artist_name = Field("artistName")
    album_name = Field("albumName")
    song_name = Field("songName")
    song_rating = Field("songRating")
    track_gain = Field("trackGain")
    track_length = Field("trackLength", 0)
    track_token = Field("trackToken")
    audio_url = Field("audioUrl")
    album_art_url = Field("albumArtUrl")
    allow_feedback = Field("allowFeedback", True)
    station_id = Field("stationId")

    album_detail_url = Field("albumDetailUrl")
    album_explore_url = Field("albumExplorerUrl")

    amazon_album_asin = Field("amazonAlbumAsin")
    amazon_album_digital_asin = Field("amazonAlbumDigitalAsin")
    amazon_album_url = Field("amazonAlbumUrl")
    amazon_song_digital_asin = Field("amazonSongDigitalAsin")

    artist_detail_url = Field("artistDetailUrl")
    artist_explore_url = Field("artistExplorerUrl")

    itunes_song_url = Field("itunesSongUrl")

    song_detail_url = Field("songDetailUrl")
    song_explore_url = Field("songExplorerUrl")

    def thumbs_up(self):
        self._api_client.add_feedback(self.track_token, True)

    def thumbs_down(self):
        self._api_client.add_feedback(self.track_token, False)

    def bookmark_song(self):
        self._api_client.add_song_bookmark(self.track_token)

    def bookmark_artist(self):
        self._api_client.add_artist_bookmark(self.track_token)

    def sleep(self):
        self._api_client.sleep_song(self.track_token)

    def get_is_playable(self):
        return URLTester(self.audio_url).is_accessible()

    @classmethod
    def from_json(cls, api_client, data):
        self = cls(api_client)

        for key, value in cls._fields.items():
            newval = data.get(value.field, value.default)

            if value.field == "audioUrl" and newval is None:
                newval = cls.get_audio_url(
                    data, api_client.default_audio_quality)

            if newval and value.formatter:
                newval = value.formatter(newval)

            setattr(self, key, newval)

        return self

    @classmethod
    def get_audio_url(cls, data,
            preferred_quality=BaseAPIClient.MED_AUDIO_QUALITY):
        """Get audio url

        Try to find audio url for specified preferred quality level, or
        next-lowest available quality url otherwise.
        """
        audio_url = None
        url_map = data.get("audioUrlMap")

        if url_map is None:
            # No audio url available (e.g. ad tokens)
            return None

        valid_audio_formats = [BaseAPIClient.HIGH_AUDIO_QUALITY,
                               BaseAPIClient.MED_AUDIO_QUALITY,
                               BaseAPIClient.LOW_AUDIO_QUALITY]

        # Only iterate over sublist, starting at preferred audio quality, or
        # from the beginning of the list if nothing is found. Ensures that the
        # bitrate used will always be the same or lower quality than was
        # specified to prevent audio from skipping for slow connections.
        i = 0
        if preferred_quality in valid_audio_formats:
            i = valid_audio_formats.index(preferred_quality)

        for quality in valid_audio_formats[i:]:
            audio_url = url_map.get(quality)

            if audio_url is not None:
                return audio_url["audioUrl"]

        return audio_url["audioUrl"] if audio_url is not None else None


class Playlist(PandoraListModel):

    __list_key__ = "items"
    __list_model__ = PlaylistItem


class Bookmark(PandoraModel):

    music_token = Field("musicToken")
    artist_name = Field("artistName")
    art_url = Field("artUrl")
    bookmark_token = Field("bookmarkToken")
    date_created = Field("dateCreated", formatter=PandoraModel.json_to_date)

    # song only
    sample_url = Field("sampleUrl")
    sample_gain = Field("sampleGain")
    album_name = Field("albumName")
    song_name = Field("songName")

    @property
    def is_song_bookmark(self):
        return self.song_name is not None

    def delete(self):
        if self.is_song_bookmark:
            self._api_client.delete_song_bookmark(self.bookmark_token)
        else:
            self._api_client.delete_artist_bookmark(self.bookmark_token)


class BookmarkList(PandoraModel):

    songs = Field("songs", formatter=Bookmark.from_json_list)
    artists = Field("artists", formatter=Bookmark.from_json_list)



#!/usr/bin/env python

try:
    from setuptools import setup
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup


setup(
    name='pydora',
    version='1.1.0',
    description='Python wrapper for Pandora API',
    long_description=open('README.rst', 'r').read(),
    author='Mike Crute',
    author_email='mcrute@gmail.com',
    url='https://github.com/mcrute/pydora',
    packages=[
        'pydora',
        'pandora',
        'pandora.models',
    ],
    install_requires=[
        'pycrypto==2.6.1',
    ],
    entry_points={
        'console_scripts': [
            'pydora = pydora.player:main',
        ],
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Multimedia :: Sound/Audio :: Players',
    ]
)



from __future__ import print_function

import sys
import termios


def input(prompt):
    try:
        return raw_input(prompt)
    except NameError:
        import builtins
        return builtins.input(prompt)


class Colors:

    def __wrap_with(code):
        @staticmethod
        def inner(text, bold=False):
            c = code
            if bold:
                c = u"1;{}".format(c)
            return u"\033[{}m{}\033[0m".format(c, text)
        return inner

    red = __wrap_with('31')
    green = __wrap_with('32')
    yellow = __wrap_with('33')
    blue = __wrap_with('34')
    magenta = __wrap_with('35')
    cyan = __wrap_with('36')
    white = __wrap_with('37')


class Screen:

    @staticmethod
    def set_echo(enabled):
        fd = sys.stdin.fileno()
        (iflag, oflag, cflag, lflag, ispeed, ospeed, cc) = \
            termios.tcgetattr(fd)

        if enabled:
            lflag |= termios.ECHO
        else:
            lflag &= ~termios.ECHO

        termios.tcsetattr(fd, termios.TCSANOW,
            [iflag, oflag, cflag, lflag, ispeed, ospeed, cc])

    @staticmethod
    def clear():
        sys.stdout.write('\x1b[2J\x1b[H')
        sys.stdout.flush()

    @staticmethod
    def print_error(msg):
        print(Colors.red(msg))

    @staticmethod
    def print_success(msg):
        print(Colors.green(msg))

    @staticmethod
    def get_integer(prompt):
        """Gather user input and convert it to an integer

        Will keep trying till the user enters an interger or until they ^C the
        program.
        """
        while True:
            try:
                return int(input(prompt).strip())
            except ValueError:
                print(Colors.red('Invaid Input!'))


def clear_screen():
    """Clear the terminal
    """
    sys.stdout.write('\x1b[2J\x1b[H')
    sys.stdout.flush()



"""
Pandora API Client

This is a reasonably complete implementation of the Pandora API. It does not
implement any of the undocumented features and does not implement most of the
account management features as they were deemed not terribly useful.

API Spec from: http://6xq.net/playground/pandora-apidoc/
Keys at: http://6xq.net/playground/pandora-apidoc/json/partners/#partners
"""
import time
import json
import base64
from Crypto.Cipher import Blowfish

try:
    from urllib.error import URLError
    from urllib.parse import urlencode
    from urllib.request import Request, urlopen
    from configparser import SafeConfigParser
except ImportError:
    from urllib import urlencode
    from urllib2 import Request, urlopen, URLError
    from ConfigParser import SafeConfigParser

from . import errors
from .errors import PandoraException


DEFAULT_API_HOST = "tuner.pandora.com/services/json/"


class APITransport(object):
    """Pandora API Transport

    The transport is responsible for speaking the low-level protocol required
    by the Pandora API. It knows about encryption, TLS and the other API
    details. Once setup the transport acts like a callable.
    """

    API_VERSION = "5"

    NO_ENCRYPT = ("auth.partnerLogin", )
    REQUIRE_TLS = ("auth.partnerLogin", "auth.userLogin",
            "station.getPlaylist", "user.createUser")

    def __init__(self, cryptor, api_host=DEFAULT_API_HOST):
        self.cryptor = cryptor
        self.api_host = api_host

        self.partner_auth_token = None
        self.user_auth_token = None

        self.partner_id = None
        self.user_id = None

        self.start_time = None
        self.server_sync_time = None

    @property
    def auth_token(self):
        if self.user_auth_token:
            return self.user_auth_token

        if self.partner_auth_token:
            return self.partner_auth_token

        return None

    @property
    def sync_time(self):
        if not self.server_sync_time:
            return None

        return int(self.server_sync_time + (time.time() - self.start_time))

    def remove_empty_values(self, data):
        return dict((k, v) for k, v in data.items() if v is not None)

    @sync_time.setter
    def sync_time(self, sync_time):
        self.server_sync_time = self.cryptor.decrypt_sync_time(sync_time)

    def _start_request(self):
        if not self.start_time:
            self.start_time = int(time.time())

    def _make_http_request(self, url, data):
        try:
            data = data.encode("utf-8")
        except AttributeError:
            pass

        req = Request(url, data, { "Content-Type": "text/plain" })
        return urlopen(req)

    def _build_url(self, method):
        query = {
            "method": method,
            "auth_token": self.auth_token,
            "partner_id": self.partner_id,
            "user_id": self.user_id,
        }

        return "{0}://{1}?{2}".format(
            "https" if method in self.REQUIRE_TLS else "http",
            self.api_host,
            urlencode(self.remove_empty_values(query)))

    def _build_data(self, method, **data):
        data["userAuthToken"] = self.user_auth_token
        data["syncTime"] = self.sync_time

        if not self.user_auth_token and self.partner_auth_token:
            data["partnerAuthToken"] = self.partner_auth_token

        data = json.dumps(self.remove_empty_values(data))

        if method not in self.NO_ENCRYPT:
            data = self.cryptor.encrypt(data)

        return data

    def _parse_response(self, result):
        result = json.loads(result.decode("utf-8"))

        if result["stat"] == "ok":
            return result["result"] if "result" in result else None
        else:
            raise PandoraException.from_code(result["code"], result["message"])

    def __call__(self, method, **data):
        self._start_request()

        url = self._build_url(method)
        data = self._build_data(method, **data)
        result = self._make_http_request(url, data)

        return self._parse_response(result.read())


class Encryptor(object):
    """Pandora Blowfish Encryptor

    The blowfish encryptor can encrypt and decrypt the relevant parts of the
    API request and response. It handles the formats that the API expects.
    """

    def __init__(self, in_key, out_key):
        self.bf_out = Blowfish.new(out_key, Blowfish.MODE_ECB)
        self.bf_in = Blowfish.new(in_key, Blowfish.MODE_ECB)

    def strip_padding(self, data):
        padding = data.find("\x00")

        if padding > 0:
            return data[:padding]

        return data

    @staticmethod
    def _decode_hex(data):
        return base64.b16decode(data.upper())

    @staticmethod
    def _encode_hex(data):
        return base64.b16encode(data).lower()

    def decrypt(self, data):
        data = self.bf_out.decrypt(self._decode_hex(data))
        return json.loads(self.strip_padding(data))

    def decrypt_sync_time(self, data):
        return int(self.bf_in.decrypt(self._decode_hex(data))[4:-2])

    def add_padding(self, data):
        plen = Blowfish.block_size - divmod(len(data), Blowfish.block_size)[1]
        return data + ("\x00" * plen)

    def encrypt(self, data):
        return self._encode_hex(self.bf_out.encrypt(self.add_padding(data)))


class URLTester(object):
    """URL Status Tester

    A utility class to make head requests to URLs and determine if they are
    currently accessible.
    """

    def __init__(self, url):
        self.url = url

    def _build_request(self):
        request = Request(self.url)
        request.get_method = lambda: "HEAD"
        return request

    def _get_status_code(self):
        request = self._build_request()
        return urlopen(request).getcode()

    def is_accessible(self):
        try:
            return self._get_status_code() == 200
        except URLError:
            return False


class BaseAPIClient(object):
    """Base Pandora API Client

    The base API client has lower level methods that are composed together to
    provide higher level functionality.
    """

    LOW_AUDIO_QUALITY = "lowQuality"
    MED_AUDIO_QUALITY = "mediumQuality"
    HIGH_AUDIO_QUALITY = "highQuality"

    def __init__(self, transport, partner_user, partner_password, device,
            default_audio_quality=MED_AUDIO_QUALITY):
        self.transport = transport
        self.partner_user = partner_user
        self.partner_password = partner_password
        self.device = device
        self.default_audio_quality = default_audio_quality
        self.username = None
        self.password = None

    @classmethod
    def from_settings_dict(cls, settings):
        enc = Encryptor(settings["DECRYPTION_KEY"], settings["ENCRYPTION_KEY"])
        trans = APITransport(enc, settings.get("API_HOST", DEFAULT_API_HOST))
        return cls(trans, settings["USERNAME"], settings["PASSWORD"],
                   settings["DEVICE"], settings["DEFAULT_AUDIO_QUALITY"])

    @classmethod
    def from_config_file(cls, path, authenticate=True):
        cfg = SafeConfigParser()
        cfg.read(path)

        self = cls.from_settings_dict(
            dict((k.upper(), v) for k, v in cfg.items("api", raw=True)))

        if authenticate and cfg.has_section("user"):
            credentials = [i[1] for i in cfg.items("user", raw=True)]
            self.login(*credentials)

        return self

    def _partner_login(self):
        partner = self.transport("auth.partnerLogin",
                username=self.partner_user,
                password=self.partner_password,
                deviceModel=self.device,
                version=self.transport.API_VERSION)

        self.transport.sync_time = partner["syncTime"]
        self.transport.partner_auth_token = partner["partnerAuthToken"]
        self.transport.partner_id = partner["partnerId"]

        return partner

    def _user_login(self, username, password):
        self.username = username
        self.password = password
        return self._authenticate()

    def _ensure_credentials_available(self):
        if not self.username or not self.password:
            raise errors.AuthenticationRequired()

    def _authenticate(self):
        self._ensure_credentials_available()

        user = self.transport("auth.userLogin",
                loginType="user",
                username=self.username,
                password=self.password,
                includePandoraOneInfo=True,
                includeSubscriptionExpiration=True,
                returnCapped=True)

        self.transport.user_id = user["userId"]
        self.transport.user_auth_token = user["userAuthToken"]

        return user

    def __call__(self, method, **kwargs):
        try:
            return self.transport(method, **kwargs)
        except errors.InvalidAuthToken:
            self._authenticate()
            return self.transport(method, **kwargs)


class APIClient(BaseAPIClient):
    """High Level Pandora API Client

    The high level API client implements the entire functional API for Pandora.
    This is what clients should actually use.
    """

    def login(self, username, password):
        self._partner_login()
        return self._user_login(username, password)

    def get_station_list(self):
        from .models.pandora import StationList

        return StationList.from_json(self,
                self("user.getStationList",
                    includeStationArtUrl=True))

    def get_station_list_checksum(self):
        return self("user.getStationListChecksum")["checksum"]

    def get_playlist(self, station_token):
        from .models.pandora import Playlist

        return Playlist.from_json(self,
                self("station.getPlaylist",
                    stationToken=station_token,
                    includeTrackLength=True))

    def get_bookmarks(self):
        from .models.pandora import BookmarkList

        return BookmarkList.from_json(self,
                self("user.getBookmarks"))

    def get_station(self, station_token):
        from .models.pandora import Station

        return Station.from_json(self,
                self("station.getStation",
                    stationToken=station_token,
                    includeExtendedAttributes=True))

    def add_artist_bookmark(self, track_token):
        return self("bookmark.addArtistBookmark",
                trackToken=track_token)

    def add_song_bookmark(self, track_token):
        return self("bookmark.addSongBookmark",
                trackToken=track_token)

    def delete_song_bookmark(self, bookmark_token):
        return self("bookmark.deleteSongBookmark",
                bookmarkToken=bookmark_token)

    def delete_artist_bookmark(self, bookmark_token):
        return self("bookmark.deleteArtistBookmark",
                bookmarkToken=bookmark_token)

    def search(self, search_text):
        return self("music.search",
                searchText=search_text)

    def add_feedback(self, track_token, positive):
        return self("station.addFeedback",
                trackToken=track_token,
                isPositive=positive)

    def add_music(self, music_token, station_token):
        return self("station.addMusic",
                musicToken=music_token,
                stationToken=station_token)

    def create_station(self, search_token=None, artist_token=None,
            track_token=None):
        kwargs = {}

        if search_token:
            kwargs = { "musicToken": search_token }
        elif artist_token:
            kwargs = { "trackToken": artist_token, "musicType": "artist" }
        elif track_token:
            kwargs = { "trackToken": track_token, "musicType": "song" }
        else:
            raise KeyError("Must pass a type of token")

        return self("station.createStation", **kwargs)

    def delete_feedback(self, feedback_id):
        return self("station.deleteFeedback",
                feedbackId=feedback_id)

    def delete_music(self, seed_id):
        return self("station.deleteMusic",
                seedId=seed_id)

    def delete_station(self, station_token):
        return self("station.deleteStation",
                stationToken=station_token)

    def get_genre_stations(self):
        return self("station.getGenreStations")

    def rename_station(self, station_token, name):
        return self("station.renameStation",
                stationToken=station_token,
                stationName=name)

    def explain_track(self, track_token):
        return self("track.explainTrack",
                trackToken=track_token)

    def set_quick_mix(self, *args):
        return self("user.setQuickMix",
                quickMixStationIds=args)

    def sleep_song(self, track_token):
        return self("user.sleepSong",
                trackToken=track_token)

    def share_station(self, station_id, station_token, *emails):
        return self("station.shareStation",
                stationId=station_id,
                stationToken=station_token,
                emails=emails)

    def transform_shared_station(self, station_token):
        return self("station.transformSharedStation",
                stationToken=station_token)

    def share_music(self, music_token, *emails):
        return self("music.shareMusic",
                musicToken=music_token,
                email=emails[0])
